/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* file.c - file descriptor state for libproxynet
 * vix 29jun94 [original]
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: file.c,v 1.6 1996/05/16 04:48:39 vixie Exp $";
#endif

#include "vixutil.h"
#include <assert.h>
#include <stdlib.h>
#include <errno.h>
#include "defs.h"

struct pcb {
	ListLink(struct pcb)
	void		*pcb;
	struct sock	cla, spa;	/* IFF this is a listener */
};

struct fd {
	ListLink(struct fd)
	int		this;
	int		other;
	enum fdType	type;
	ListHead(	pcbs, struct pcb)
	int		flags;
#define	FD_WILDPORT	0x0001
};

static struct {
	ListHead(	fds, struct fd)
} State;

/*----------------------------------------------------------- private ---*/

#ifdef DEBUG_local
static void
FDlist(FILE *f) {
	register struct fd *hp;

	ListEach(&State.fds, hp) {
		register struct pcb *pcbp;

		fprintf(f, "FDlist[%d] (%d %d 0x%x)",
			hp->this, hp->other, (int)hp->type, hp->flags);
		ListEach(&hp->pcbs, pcbp)
			fprintf(f, " 0x%x", pcbp->pcb);
		fprintf(f, "\n");
	}
}
#else
# define FDlist(f)
#endif

static struct fd *
FDfind(register int this) {
	register struct fd *hp;

	ListEach(&State.fds, hp)
		if (hp->this == this)
			return (hp);
	errno = ENOENT;
	return (NULL);
}

static struct fd *
FDnew(int this) {
	register struct fd *hp = ZALLOC(struct fd, ZAL_ZERO);

	if (!hp) {
		errno = ENOMEM;
	} else {
		hp->this = this;
		hp->other = -1;
		hp->type = fdTnone;
		hp->flags = 0;
		ListInit(&hp->pcbs);
		ListAdd(&State.fds, hp);
		FDlist(stderr);
	}
	return (hp);
}

static struct pcb *
PCBfind(struct fd *hp, void *pcb) {
	register struct pcb *pcbp;

	ListEach(&hp->pcbs, pcbp)
		if (pcbp->pcb == pcb)
			return (pcbp);
	errno = ENOENT;
	return (NULL);
}

static struct pcb *
PCBnew(struct fd *hp, void *pcb) {
	struct pcb *pcbp = ZALLOC(struct pcb, ZAL_ZERO);

	if (!pcbp) {
		errno = ENOMEM;
	} else {
		ListAdd(&hp->pcbs, pcbp);
		pcbp->pcb = pcb;
	}
	return (pcbp);
}

/*----------------------------------------------------------- public ---*/

int
proxfdSetT(int this, enum fdType type) {
	register struct fd *hp;

	assert(type != fdTnone);
	proxyTr(PROXTRC_FILEDES, "proxfdSetT(%d, %d)\n", this, (int)type);
	if ((hp = FDfind(this)) != NULL || (hp = FDnew(this)) != NULL)
		hp->type = type;
	return (hp ? 0 : -1);
}

enum fdType
proxfdGetT(int this) {
	register struct fd *hp = FDfind(this);
	register enum fdType type = hp ? hp->type : fdTerror;

	proxyTr(PROXTRC_FILEDES, "proxfdGetT(%d) -> %d\n", this, (int)type);
	return (type);
}

int
proxfdEach(const proxfdEachF func, void *uap) {
	register struct fd *hp;
	register int result = 0;

	ListEach(&State.fds, hp) {
		result = (*func)(hp->this, uap);
		if (result != 0)
			break;
	}
	return (result);
}

/*
 * PCB hints are a list of {PCB,cla,spa}'s which have been bound to this FD.
 * Attempting to add a duplicate hint will be silently ignored.
 */

int
proxfdSetL(int this, void *pcb,
	   const struct sock *cla, const struct sock *spa)
{
	register struct fd *hp;

	if ((hp = FDfind(this)) != NULL || (hp = FDnew(this)) != NULL) {
		if (hp->type == fdTnone || hp->type == fdTlisten) {
			struct pcb *pcbp;

			if ((pcbp = PCBfind(hp, pcb)) != NULL ||
			    (pcbp = PCBnew(hp, pcb)) != NULL) {
				hp->type = fdTlisten;
				pcbp->cla = *cla;
				pcbp->spa = *spa;
				return (0);
			}
		} else {
			errno = EFTYPE;
		}
	}
	return (-1);
}

int
proxfdGetL(int this, void *pcb, struct sock *cla, struct sock *spa) {
	struct fd *hp = FDfind(this);

	if (hp) {
		if (hp->type == fdTlisten) {
			struct pcb *pcbp = PCBfind(hp, pcb);

			if (pcbp) {
				*cla = pcbp->cla;
				*spa = pcbp->spa;
				return (0);
			}
		} else {
			errno = EFTYPE;
		}
	}
	return (-1);
}

int
proxfdSetP(int this, void *pcb) {
	register struct fd *hp;

	proxyTr(PROXTRC_FILEDES, "proxfdSetP(%d, 0x%x)\n", this, pcb);
	if ((hp = FDfind(this)) != NULL || (hp = FDnew(this)) != NULL) {
		struct pcb *pcbp;

		if ((pcbp = PCBfind(hp, pcb)) != NULL ||
		    (pcbp = PCBnew(hp, pcb)) != NULL) {
			pcbp->pcb = pcb;
			return (0);
		}
	}
	return (-1);
}

/*
 * Here we iterate by calling the specified function until we run out
 * of PCB's or get back a non-zero result from that function.  We return
 * that result, or 0 if we hit the end of the list naturally.
 */

int
proxfdEachP(int this, proxfdEachPF func, void *uap) {
	register struct fd *hp;
	register struct pcb *pcbp;
	register int result = 0;

	if (!(hp = FDfind(this))) {
		proxyTr(PROXTRC_FILEDES, "proxfdEachP(%d): ENOENT\n", this);
		errno = ENOENT;
		result = -1;
		goto done;
	}
	ListEach(&hp->pcbs, pcbp) {
		result = (*func)(pcbp->pcb, this, uap);
		proxyTr(PROXTRC_FILEDES, "proxfdEachP(%d) [0x%x] -> %d\n",
			this, pcbp->pcb, result);
		if (result != 0)
			break;
	}
 done:
	return (result);
}

int
proxfdSet(int other, int this) {
	struct f  { ListLink(struct f); int f; }  *f, *fn, *fp;
	ListHeadV(fs, struct f);
	register struct fd *hp, *hp2;
	int s, result, save;

	proxyTr(PROXTRC_FILEDES, "proxfdSet(%d, %d)\n", other, this);
	ListInit(&fs);

	/* It is not reasonable to have a file descriptor inherit its
	 * proxiness from more than one other file descriptor, or to
	 * change its association without first being closed (or bound
	 * to a new address, which is a case we don't catch right now).
	 */
	if ((hp = FDfind(this)) != NULL && hp->other != -1) {
		errno = EALREADY;
		result = -1;
		goto done;
	}
	if (!hp && !(hp = FDnew(this))) {
		errno = ENOMEM;
		result = -1;
		goto done;
	}

	/* Detect and disallow association loops.  If our proposed
	 * "other" is the same as some previous "this" which would
	 * lead to us via its "other" chain, we have an error.
	 *
	 * Sensible as it may seem to use "fd_set"'s for this, they
	 * can be bigger in the application's compile than in ours
	 * (see select(2)).
	 */
	for (s = other;  (hp2 = FDfind(s)) != NULL;  s = hp2->other) {
		f = ZALLOC(struct f, ZAL_ZERO);
		if (f == NULL) {
			errno = ENOMEM;
			result = -1;
			goto done;
		}
		f->f = hp2->other;
		ListAdd(&fs, f);
	}
	ListEach(&fs, f) {
		if (f->f == this) {
			errno = ELOOP;
			result = -1;
			goto done;
		}
	}

	/* Looks like everything will be ok.  Make it happen.
	 */
	hp->other = other;
	result = 0;

 done:
	save = errno;
	fp = NULL;
	ListEach2(&fs, f, fn) {
		ListUnlink(&fs, f, fp);		/* pedantic */
		free(f);
	}
	errno = save;
	return (result);
}

int
proxfdGet(int this) {
	register struct fd *hp = FDfind(this);

	proxyTr(PROXTRC_FILEDES, "proxfdGet(%d) -> %d\n",
		this, (hp ? hp->other : -1));
	if (hp)
		return (hp->other);
	return (-1);
}

int
proxfdSetW(int this, int flag) {
	register struct fd *hp;

	proxyTr(PROXTRC_FILEDES, "proxfdSetW(%d, %d)\n", this, flag);
	if ((hp = FDfind(this)) != NULL || (hp = FDnew(this)) != NULL) {
		if (flag)
			hp->flags |= FD_WILDPORT;
		else
			hp->flags &= ~FD_WILDPORT;
		return (0);
	}
	return (-1);
}

int
proxfdGetW(int this) {
	register struct fd *hp = FDfind(this);

	proxyTr(PROXTRC_FILEDES, "proxfdGetW(%d) -> %d\n",
		this, (hp ? ((hp->flags & FD_WILDPORT) != 0) : -1));
	if (hp)
		return ((hp->flags & FD_WILDPORT) != 0);
	return (-1);
}

int
proxfdClr(register int this) {
	register struct fd *hp, *hpn, *hpp;
	register struct pcb *pcbp, *pcbpn, *pcbpp;
	register int count = 0;

	proxyTr(PROXTRC_FILEDES, "proxfdClr(%d)\n", this);
	hpp = NULL;
	ListEach2(&State.fds, hp, hpn) {
		if (hp->this == this) {
			pcbpp = NULL;
			ListEach2(&hp->pcbs, pcbp, pcbpn) {
				ListUnlink(&hp->pcbs, pcbp, pcbpp);
				free(pcbp);
			}
			ListUnlink(&State.fds, hp, hpp);
			free(hp);
			count++;
		} else {
			hpp = hp;
		}
	}
	FDlist(stderr);
	return (count);
}
