/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* proxy.c - transaction layer (contains config data, calls service layer)
 * vix 02feb94 [original]
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: proxy.c,v 1.10 1996/05/16 04:48:39 vixie Exp $";
#endif

#include "vixutil.h"

#include <sys/stat.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include <stdarg.h>

#include "defs.h"
#include "pathnames.h"
#include "messages.h"

struct conf {
#if PROXY_RESET
	time_t		mtime;
#endif
	ListHead(	proxies, struct proxy)
	int		flags;
#define CONF_INIT	0x0001
};

struct proxy {
	ListLink(struct proxy)
	ListHead(	addrs,	struct proxAdr)
	ListHead(	names,	struct proxNam)
	ListHead(	psrvs,	struct addr)
	ListHead(	nsrvs,	struct addr)
};

struct proxAdr {
	ListLink(struct proxAdr)
	struct sock	addr,
			mask;
	u_short		flags;
#define	PROXADR_NOT	0x0001
#define	PROXADR_MASKED	0x0002
};

enum pnAssoc {
	pnExact,	/* exact match (no "*") */
	pnLeft,		/* partial left match ("*" on right) */
	pnRight,	/* partial right match ("*" on left) */
	pnWild		/* wildcard match (just a "*") */
};

struct proxNam {
	ListLink(struct proxNam)
	char		*name;
	enum pnAssoc	assoc;
	short		len;
	u_short		flags;
#define	PROXNAM_NOT	0x0001
};

struct addr {
	ListLink(struct addr)
	struct sock	addr;
	u_short		flags;
#define	ADDR_WEIGHT	0x0001
	struct interval	reopen,
			keepalive;
};

static const char	Path[] = _PATH_PROXYTCP_CONF,
			TraceEnv[] = PROXYNET_TRACE,
			NullTokenStr[] = "-";

static struct conf	Conf;


static FILE *		proxyTrans1 __P((struct addr *psp, const char *req));


#ifdef DEBUG
static void
printConf(struct conf *c) {
	proxyTr(PROXTRC_INIT, "+++printConf+++\n");
	if (c->flags & CONF_INIT) {
		struct proxy *pp;

		ListEach(&c->proxies, pp) {
			struct proxAdr *pap;
			struct proxNam *pnp;
			struct addr *ap;

			proxyTr(PROXTRC_INIT, "proxy\t");
			ListEach(&pp->addrs, pap) {
				proxyTr(PROXTRC_INIT, "%s%s",
					(pap->flags & PROXADR_NOT) ?"!" :"",
					sock_ntoa(&pap->addr));
				if (pap->flags & PROXADR_MASKED)
					proxyTr(PROXTRC_INIT, "&%s",
						sock_ntoa(&pap->mask));
				if (!ListLast_p(pap))
					proxyTr(PROXTRC_INIT, ",\n\t\t");
			}
			proxyTr(PROXTRC_INIT, "%s\n\t",
				ListEmpty_p(&pp->addrs) ?NullTokenStr :"");
			ListEach(&pp->names, pnp) {
				proxyTr(PROXTRC_INIT, "%s%s%s%s%s",
					(pnp->flags & PROXNAM_NOT) ?"!" :"",
					(pnp->assoc == pnRight ||
					 pnp->assoc == pnWild) ?"*" :"",
					pnp->name,
					(pnp->assoc == pnLeft) ?"*" :"",
					ListLast_p(pnp) ?"" :",");
			}
			proxyTr(PROXTRC_INIT, "%s\n\t",
				ListEmpty_p(&pp->names) ?NullTokenStr :"");
			ListEach(&pp->psrvs, ap) {
				proxyTr(PROXTRC_INIT, "%s%s",
					sock_ntoa(&ap->addr),
					ListLast_p(ap) ?"" :",");
			}
			proxyTr(PROXTRC_INIT, "%s\n\t",
				ListEmpty_p(&pp->psrvs) ?NullTokenStr :"");
			ListEach(&pp->nsrvs, ap) {
				proxyTr(PROXTRC_INIT, "%s%s",
					sock_ntoa(&ap->addr),
					ListLast_p(ap) ?"" :",");
			}
			proxyTr(PROXTRC_INIT, "%s;\n",
				ListEmpty_p(&pp->nsrvs) ?NullTokenStr :"");
		}
	}
	proxyTr(PROXTRC_INIT, "---printConf---\n");
}
#endif /*DEBUG*/

/* XXX:	this should probably become part of libvixutil/sock_*.c, since it
 *	should become possible to specify just "&" DOTTEDQUAD "/" PORT rather
 *	than having the mask need to be a fully glorified sockspec.  this could
 *	also be done by making a sock_aton2() which took a template which
 *	was used for defaults.  then it would be "&//" DOTTEDQUAD ["/" PORT]
 *	which is easier than specifying the whole thing, though harder to
 *	document.
 */
static const char *
doMask(char *mask, struct sock *ma, struct sock *sa) {
	if (!sock_aton(mask, ma))
		return (proxmsg_BadMask);
	if (ma->type != sa->type)
		return (proxmsg_MismatchTyp);
	if (sock_af(ma) != sock_af(sa))
		return (proxmsg_MismatchFam);
	switch (sock_af(ma)) {
	case AF_INET:
		if (((struct sockaddr_in *)&sa->addr)->sin_addr.s_addr &
		    ~((struct sockaddr_in *)&ma->addr)->sin_addr.s_addr)
			return (proxmsg_AddrExceeds);
		if (ntohs(((struct sockaddr_in *)&ma->addr)->sin_port) != 0)
			return (proxmsg_BadMask);
		break;
	default:
		return (proxmsg_NoMaskProto);
	}
	return (NULL);
}


#if PROXY_RESET
static void
proxyReset() {
	struct proxy *pp, *pt;

	ListEach2(&Conf.proxies, pp, pt) {
		struct proxAdr *ap, *at;
		struct proxNam *dp, *dt;
		struct addr *p, *t;

		ListEach2(&pp->addrs, ap, at)
			free(ap);
		ListEach2(&pp->names, dp, dt) {
			free(dp->name);
			free(dp);
		}
		ListEach2(&pp->psrvs, p, t)
			free(p);
		ListEach2(&pp->nsrvs, p, t)
			free(p);
		free(pp);
	}
	memset(&Conf, 0, sizeof Conf);
}
#endif


static int
proxyLoadConf() {
	static const char comstr[] = "#";
	char tok[MEDIUM], ptok[MEDIUM];
	const char *msg;
	int sep, result;
	FILE *f;

#define	nexttok(sepStr) (strcpy(ptok, tok), \
			 fgettok(f, tok, MEDIUM, sepStr, comstr))

	if (!(f = fopen(Path, "r"))) {
		result = -1;
		goto done;
	}
	tok[0] = '\0';
	ptok[0] = '\0';
	while (0 < (sep = nexttok(" \n"))) {
		if (!strcmp(tok, "proxy")) {
			struct proxy *np;

			if (!(np = ZALLOC(struct proxy, ZAL_ZERO))) {
				msg = proxmsg_OutOfMem;
				goto err;
			}
			/*------- addresses --------*/
			while (0 < (sep = nexttok(" \n,"))) {
				if (strcmp(tok, NullTokenStr)) {
					char *cp = tok, *mask;
					int not = 0;
					struct sock sa, ma;
					struct proxAdr *npa;

					if (*cp == '!') {
						not++;
						cp++;
					}
					if (mask = strchr(tok, '&'))
						*mask++ = '\0';
					if (!sock_aton(cp, &sa)) {
						msg = proxmsg_BadAddress;
						goto err;
					}
					if (mask) {
						msg = doMask(mask, &ma, &sa);
						if (msg)
							goto err;
					}
					npa = ZALLOC(struct proxAdr, ZAL_ZERO);
					if (!npa) {
						msg = proxmsg_OutOfMem;
						goto err;
					}
					npa->addr = sa;
					npa->mask = ma;
					if (not)
						npa->flags |= PROXADR_NOT;
					if (mask)
						npa->flags |= PROXADR_MASKED;
					ListAdd(&np->addrs, npa);
				}
				if (sep != ',')
					break;
			}
			if (sep == EOF) {
				msg = proxmsg_EndOfFile;
				goto err;
			}
			/*------- names --------*/
			while (0 < (sep = nexttok(" \n,"))) {
				if (strcmp(tok, NullTokenStr)) {
					int len, not, star_l, star_r;
					struct proxNam *npn;
					char *cp = tok;

					not = 0;
					if (*cp == '!') {
						not++;
						cp++;
					}
					star_l = 0;
					if (*cp == '*') {
						star_l++;
						cp++;
					}
					len = strlen(cp);
					if (!len && !star_l) {
						msg = proxmsg_NullName;
						goto err;
					}
					star_r = 0;
					if (len && cp[len-1] == '*') {
						star_r++;
						cp[--len] = '\0';
					}
					if (strchr(cp, '*') ||
					    (star_l && star_r)) {
						msg = proxmsg_BadNamAssoc;
						goto err;
					}
					npn = ZALLOC(struct proxNam, ZAL_ZERO);
					if (!npn) {
						msg = proxmsg_OutOfMem;
						goto err;
					}
					npn->name = strdup(cp);
					npn->len = len;
					if (star_l)
						if (len)
							npn->assoc = pnRight;
						else
							npn->assoc = pnWild;
					else
						if (star_r)
							npn->assoc = pnLeft;
						else
							npn->assoc = pnExact;
					if (not)
						npn->flags |= PROXNAM_NOT;
					ListAdd(&np->names, npn);
				}
				if (sep != ',')
					break;
			}
			if (sep == EOF) {
				msg = proxmsg_EndOfFile;
				goto err;
			}
			/*----- proxy servers -----*/
			while (0 < (sep = nexttok(" \n,"))) {
				struct addr *na;
				struct sock sa;

				if (strcmp(tok, NullTokenStr)) {
					if (!sock_aton(tok, &sa)) {
						msg = proxmsg_BadProxSrvA;
						goto err;
					}
					na = ZALLOC(struct addr, ZAL_ZERO);
					if (!na) {
						msg = proxmsg_OutOfMem;
						goto err;
					}
					na->addr = sa;
					ListAdd(&np->psrvs, na);
				}
				if (sep != ',')
					break;
			}
			/*----- name servers -----*/
			while (0 < (sep = nexttok(" \n,;"))) {
				struct addr *na;
				struct sock sa;

				if (strcmp(tok, NullTokenStr)) {
					if (!sock_aton(tok, &sa) ||
					    sock_af(&sa) != AF_INET ||
					    !sock_isdgram(&sa)) {
						msg = proxmsg_BadNameSrvA;
						goto err;
					}
					na = ZALLOC(struct addr, ZAL_ZERO);
					if (!na) {
						msg = proxmsg_OutOfMem;
						goto err;
					}
					na->addr = sa;
					ListAdd(&np->nsrvs, na);
				}
				if (sep != ',')
					break;
			}
			/*----- final -----*/
			ListAdd(&Conf.proxies, np);
		} else {
			msg = proxmsg_UnrecogVerb;
			goto err;
		}
	}
	result = 0;
	goto done;
 err:
	syslog(LOG_WARNING, proxmsg_ConfigError, msg, tok, ptok);
	result = -1;
	errno = EFTYPE;		/* XXX - got a better suggestion? */
 done:
	if (f)
		fclose(f);
	return (result);
}


int
proxyInit() {
	static int Entered;
	int result;
#if PROXY_RESET
	struct stat statbuf;
#endif

	if (Entered) {
		errno = EALREADY;
		return (-1);
	}
	Entered++;

	proxresInit();
	proxsrvInit();
#if PROXY_RESET
	if (0 > stat(Path, &statbuf)) {
		result = -1;
		goto done;
	}
#endif
	if (Conf.flags & CONF_INIT) {
#if PROXY_RESET
		if (Conf.mtime == statbuf.st_mtime) {
			result = 0;
			goto done;
		}
		proxresReset();		/* invalidate its local cache */
		proxyReset();
#else
		result = 0;
		goto done;
#endif
	}
	result = proxyLoadConf();
	if (result == 0) {
#if PROXY_RESET
		Conf.mtime = statbuf.st_mtime;
#endif
		Conf.flags |= CONF_INIT;
#ifdef DEBUG
		printConf(&Conf);
#endif
	}
 done:
	Entered--;
	return (result);
}


static struct proxAdr *
addrOnList(const struct proxy *pp, const struct sock *addr) {
	/* XXX - we only handle IP right now. */
	if (sock_af(addr) == AF_INET) {
		struct proxAdr *pap;
		struct sockaddr_in *in = (struct sockaddr_in *)&addr->addr;

		ListEach(&pp->addrs, pap) {
			register struct sockaddr_in *ap, *mp;
			struct in_addr srch, list;
			int srch_wildaddr;

			if (sock_af(&pap->addr) != AF_INET)
				continue;

			ap = (struct sockaddr_in *)&pap->addr.addr;
			mp = (struct sockaddr_in *)&pap->mask.addr;
			srch = in->sin_addr;
			list = ap->sin_addr;
			if (pap->flags & PROXADR_MASKED) {
				srch.s_addr &= mp->sin_addr.s_addr;
				list.s_addr &= mp->sin_addr.s_addr;
			}

/* given ip/tcp/0.0.0.0/119
 * then ip/tcp/0.0.0.0/0	NO MATCH
 *	ip/tcp/10.0.0.0/0	NO MATCH
 *	ip/tcp/0.0.0.0/119	MATCH
 *	ip/tcp/10.0.0.0/119	MATCH
 *	ip/tcp/10.0.0.0/25	NO MATCH
 *
 * given ip/tcp/10.1.0.1/0
 * then ip/tcp/0.0.0.0/0	MATCH
 *	ip/tcp/10.0.0.0/0	MATCH		(assume 255.0.0.0 mask)
 *	ip/tcp/0.0.0.0/119	NO MATCH	(assume 0.0.0.0 mask)
 *	ip/tcp/10.0.0.0/119	NO MATCH
 *
 * given ip/tcp/10.1.0.1/119
 * then ip/tcp/0.0.0.0/0	NO MATCH
 *	ip/tcp/10.0.0.0/0	NO MATCH	(assume 255.0.0.0 mask)
 *	ip/tcp/0.0.0.0/119	MATCH		(assume 0.0.0.0 mask)
 *	ip/tcp/10.0.0.0/119	MATCH
 */

			srch_wildaddr = IN_WILD_ADDR(*in);
			if (!srch_wildaddr && srch.s_addr != list.s_addr)
				continue;
			if (IN_WILD_PORT(*ap)) {
				if (srch_wildaddr)
					continue;
			} else if (in->sin_port != ap->sin_port)
				continue;

			/* match */
			return (pap);
		}
	}
	return (NULL);
}


static struct proxNam *
nameOnList(const struct proxy *pp, const char *name) {
	int len = strlen(name);
	struct proxNam *pnp;

	ListEach(&pp->names, pnp) {
		int match;

		switch (pnp->assoc) {
		case pnWild:
			match = 1;
			break;
		case pnExact:
			match = !strcasecmp(pnp->name, name);
			break;
		case pnLeft:
			match = (len >= pnp->len) &&
				!strncasecmp(pnp->name, name, pnp->len);
			break;
		case pnRight:
			match = (len >= pnp->len) &&
				!strcasecmp(pnp->name, name + len - pnp->len);
			break;
		default:
			/* XXX:	syslog? ASSERT? */
			match = 0;
			break;
		}

		if (match)
			return (pnp);
	}
	return (NULL);
}


int
proxyProxsrvs_p(void *pcb) {
	struct proxy *pp = (struct proxy *) pcb;

	return (!ListEmpty_p(&pp->psrvs));
}


int
proxyNamesrvs_p(void *pcb) {
	struct proxy *pp = (struct proxy *) pcb;

	return (!ListEmpty_p(&pp->nsrvs));
}


void *
proxyThisAddr(const struct sock *addr) {
	struct proxy *pp;
	struct proxAdr *pap;
	void *result;

	/*
	 * We only handle IP streams right now.
	 */
	if (!sock_isstream(addr) || sock_af(addr) != AF_INET)
		return (NULL);

	/*
	 * Attempt (re-)initialization.
	 */
	if (0 > proxyInit())
		return (NULL);

	/*
	 * Search proxies.
	 */
	result = NULL;
	ListEach(&Conf.proxies, pp) {
		if (pap = addrOnList(pp, addr)) {
			/*
			 * Matched.
			 */
			if (!(pap->flags & PROXADR_NOT))
				result = pp;	/* "pcb" */
			break;
		}
	}
	proxyTr(PROXTRC_LOOKUPS, "proxyThisAddr(%s): 0x%lx\n",
		sock_ntoa(addr), result);
	return (result);
}


void *
proxyThisName(const char *name) {
	struct proxy *pp;
	struct proxNam *pnp;
	void *result;

	/*
	 * Attempt (re-)initialization.
	 */
	if (0 > proxyInit())
		return (NULL);

	/*
	 * Search proxies.
	 */
	result = NULL;
	ListEach(&Conf.proxies, pp) {
		if (pnp = nameOnList(pp, name)) {
			/*
			 * Matched.
			 */
			if (!(pnp->flags & PROXNAM_NOT))
				result = pp;	/* "pcb" */
			break;
		}
	}
	proxyTr(PROXTRC_LOOKUPS, "proxyThisName(%s): 0x%lx\n", name, result);
	return (result);
}


void *
proxyAddrIsProxy(const struct sock *addr) {
	struct proxy *pp;
	struct addr *psp;
	struct sock t;
	void *result;

	/*
	 * We only handle IP streams right now. (XXX)
	 */
	if (!sock_isstream(addr) || sock_af(addr) != AF_INET)
		return (NULL);

	/*
	 * Attempt (re-)initialization.
	 */
	if (0 > proxyInit())
		return (NULL);

	/*
	 * Search server addresses of all proxies.
	 */
	result = NULL;
	t = *addr;
	ListEach(&Conf.proxies, pp) {
		ListEach(&pp->psrvs, psp) {
			((struct sockaddr_in *)&t.addr)->sin_port =
			     ((struct sockaddr_in *)&psp->addr.addr)->sin_port;
			if (sock_same(&t, &psp->addr)) {
				result = pp;	/* "pcb" */
				break;
			}
		}
		if (result)
			break;
	}
	proxyTr(PROXTRC_LOOKUPS, "proxyAddrIsProxy(%s): 0x%lx\n",
		sock_ntoa(addr), result);
	return (result);
}


int
proxyEachSvcMatch(const struct sock *addr, const proxEachF func, int s) {
	struct proxy *pp;
	struct addr *psp;
	struct sock spa;
	int count;

	if (!sock_isstream(addr) || sock_af(addr) != AF_INET) {	/* XXX - IP */
		errno = EAFNOSUPPORT;
		return (-1);
	}

	if (0 > proxyInit())
		return (-1);

	/*
	 * Build temporary sock for service proxy address.
	 */
	spa = *addr;
	((struct sockaddr_in *)&spa.addr)->sin_addr.s_addr = INADDR_ANY;

	/*
	 * Search target addresses of all proxies.
	 */
	count = 0;
	ListEach(&Conf.proxies, pp) {
		if (!addrOnList(pp, &spa))
			continue;
		ListEach(&pp->psrvs, psp) {
			FILE *pfp = proxsrvOpenF(&psp->addr);	/*cache only*/
			struct sock cla;

			if (!pfp || sock_name(fileno(pfp), &cla) == -1)
				continue;
			((struct sockaddr_in *)&cla.addr)->sin_port =
				((struct sockaddr_in *)&addr->addr)->sin_port;
			(void) (*func)(&cla, &spa, pp, s);
			count++;
		}
	}

	return (count);
}


/*
 * Implement round robin, get/reuse connection to least-weighted server of
 * proxy (psp), send request (req), return open FILE* or NULL on errors.
 */
FILE *
proxyTransaction(const void *pcb, const char *req) {
	struct proxy *pp = (struct proxy *)pcb;
	struct addr *best_psp, *psp;
	u_int best_weight, cur_weight;
	int nps, tries;
	FILE *fp;

	/*
	 * We should not be called unless there are proxy servers defined.
	 */
	if (!proxyProxsrvs_p(pcb)) {
		errno = EDESTADDRREQ;
		return (NULL);
	}

	/*
	 * Implement round-robin.
	 */
	tries = 0;
 again:
	/*
	 * Pass 1: find the lowest weight and first PSP that has it.
	 */
	best_weight = 2;
	best_psp = NULL;
	nps = 0;
	ListEach(&pp->psrvs, psp) {
		cur_weight = (psp->flags & ADDR_WEIGHT) ? 1 : 0;
		if (cur_weight < best_weight) {
			best_weight = cur_weight;
			best_psp = psp;
		}
		nps++;
	}

	/*
	 * Pass 2: clear all weights if all are set
	 */
	if (best_weight > 0)
		ListEach(&pp->psrvs, psp)
			psp->flags &= ~ADDR_WEIGHT;

	/*
	 * Use best PSP.
	 */
	psp = best_psp;
	psp->flags |= ADDR_WEIGHT;

	fp = proxyTrans1(psp, req);
	if (!fp) {
		int save = errno;

		if (++tries < nps)
			goto again;
		syslog(LOG_WARNING, proxmsg_NoServers);
		errno = save;
	}
	return (fp);
}


void
proxyKeepAlive(const void *pcb) {
	struct proxy *pp = (struct proxy *)pcb;
	struct addr *psp;

	ListEach(&pp->psrvs, psp) {
		FILE *fp;

		if (interval_overdue_p(&psp->keepalive)) {
			fp = proxyTrans1(psp, "NOOP");
			if (fp) {
				char buf[ARPA_MAXSTR];
				int code;

				while (!proxsrvResp(fp, buf, sizeof buf,
						    &code))
					NULL;
				interval_reset(&psp->keepalive, RETRY_DEFAULT);
			}
		}
	}
}


/*
 * Send given request (req) to all servers of proxy (psp), return zilch.
 * Used for LSTN, which doesn't care whether they succeed since it is
 * supposed to lead to keepalives later.
 */
void
proxyTransEach(const void *pcb, const char *req) {
	struct proxy *pp = (struct proxy *)pcb;
	struct addr *psp;

	ListEach(&pp->psrvs, psp) {
		FILE *fp = proxyTrans1(psp, req);

		if (fp) {
			char buf[ARPA_MAXSTR];
			int code;

			while (!proxsrvResp(fp, buf, sizeof buf, &code))
				NULL;
		}
	}
}


/*
 * Get/reuse connection to designated server (psp), send request (req),
 * return open FILE* or NULL on errors.
 */
static FILE *
proxyTrans1(struct addr *psp, const char *req) {
	FILE *fp;

	/*
	 * Get (maybe reuse) a connection to this proxy server.
	 * XXX - IP specific.
	 */
	if (sock_af(&psp->addr) == AF_INET &&
	    IN_WILD_PORT(*(struct sockaddr_in *)&psp->addr.addr)) {
		/* XXX - should use getservbyname() and then default. */
		((struct sockaddr_in *)&psp->addr.addr)->sin_port =
			htons(PROXYNET_PORT);
	}
	fp = proxsrvOpenF(&psp->addr);
	if (!fp) {
		/* It's not currently open, don't retry too soon. */
		if (interval_overdue_p(&psp->reopen)) {
			fp = proxsrvOpen(&psp->addr);
			if (!fp) {
				interval_reset(&psp->reopen,
					       (errno == ETIMEDOUT)
					       ? RETRY_TIMEOUT
					       : RETRY_DEFAULT);
			}
		}
	}
	/* If it's still not open after all of that, we're outta here. */
	if (!fp)
		return (NULL);

	/*
	 * Send command, return open FP for response.
	 */
	proxyTr(PROXTRC_PROTO, ">>> %s\n", req);
	fprintf(fp, "%s\n", req);
	if (ferror(fp)) {
		/* Was up, now down. */
		proxsrvClose(fp);
		interval_reset(&psp->reopen, RETRY_DEFAULT);
		return (NULL);
	}
	return (fp);
}


/*
 * Fill in name servers in given "res" struct (vp1) from proxy's (vp2) list.
 */
void
proxyOverrideNS(void *vp1, void *vp2) {
	struct __res_state *rp = (struct __res_state *)vp1;
	struct proxy *pp = (struct proxy *)vp2;
	struct addr *pap;
	int n = 0;

	ListEach(&pp->nsrvs, pap) {
		rp->nsaddr_list[n] = *(struct sockaddr_in *)&pap->addr.addr;
		if (++n > MAXNS)
			break;
	}
	rp->nscount = n;
}


#ifdef DEBUG
void
proxyTr(unsigned int mask, const char *fmt, ...) {
	static unsigned int initFlag, traceMask;
	static FILE *traceFile;
	va_list	ap;

	if (!initFlag) {
		char *v, *f, c;

		v = getenv(TraceEnv);
		if (!v) {
			traceMask = 0;
		} else {
			f = strchr(v, ':');
			if (!f) {
				traceFile = stderr;
			} else {
				if (!(traceFile = fopen(f+1, "a")))
					traceFile = fopen(f+1, "w");
				if (!traceFile) {
					perror(traceFile);
					traceFile = stderr;
				}
				c = *f;
				*f = '\0';
			}
			traceMask = atoi(v);
			if (f)
				*f = c;
		}
		initFlag++;
	}
	va_start(ap, fmt);
	if ((traceMask & mask) && traceFile) {
		vfprintf(traceFile, fmt, ap);
		fflush(traceFile);
	}
	va_end(ap);
}

void
proxyLog(const char *message, ...) {
	va_list ap;

	va_start(ap, message);
	vsyslog(LOG_DEBUG, message, ap);
	va_end(ap);
}
#endif /*DEBUG*/
