/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* service.c - server communication layer of libproxynet
 * vix 16mar94 [original]
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: service.c,v 1.7 1996/05/16 04:48:39 vixie Exp $";
#endif

#include "vixutil.h"

#include <sys/time.h>
#include <sys/fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <setjmp.h>
#include <syslog.h>

#include "defs.h"
#include "pathnames.h"

struct proxSrv {
	ListLink(struct proxSrv);
	struct sock	addr;
	FILE		*fp;
};

static struct {
	ListHead(	srv, struct proxSrv)
	pid_t		pid;
} State;

static jmp_buf timejump;
static void sigalrm(int junk) {
	longjmp(timejump, 1);
}

void
proxsrvInit() {
	pid_t pid = getpid();

	/*
	 * This catches sendmail's fork().  Having children intersperse their
	 * transactions to the same proxy connection is quite confusing.
	 */
	if (State.pid && (State.pid != pid)) {
		proxyTr(PROXTRC_INIT,
			"pid differs (%lu != %lu); closing ctls\n",
			(u_long)State.pid, (u_long)pid);
		proxsrvCloseAll();
	}
	State.pid = pid;
}

/*
 * Search for an open connection to this address and return it;
 * If no match, return NULL.  This implements connection caching.
 */
static struct proxSrv *
proxsrvOpenP(const struct sock *addr) {
	register struct proxSrv *psp;

	ListEach(&State.srv, psp)
		if (sock_same(&psp->addr, addr))
			return (psp);
	return (NULL);
}

/*
 * Search for an open connection to this address and return its FILE *.
 * If no match, return NULL.  This is a window into the connection cache.
 */
FILE *
proxsrvOpenF(const struct sock *addr) {
	register struct proxSrv *psp = proxsrvOpenP(addr);

	if (psp)
		return (psp->fp);
	return (NULL);
}

/*
 * Get a connection to the given address, either from the cache or by
 * making a new one.  Returns NULL only if the connection failed.
 */
FILE *
proxsrvOpen(const struct sock *addr) {
	struct itimerval timeout, oldtimeout;
	struct proxSrv *psp;
	char buf[ARPA_MAXSTR];
	sig_t oldalrm;
	sigset_t set, oset;
	int fd, code;
	FILE *conn;

	proxsrvInit();

	psp = proxsrvOpenP(addr);
	if (psp) {
		if (!feof(psp->fp) && !ferror(psp->fp))
			return (psp->fp);
		proxsrvClose(psp->fp);	/* (deletes/unlinks *psp) */
		syslog(LOG_WARNING, "EOF or Error, closing proxy server %s",
		       sock_ntoa(&psp->addr));
	}

	/*
	 * Open a new connection.
	 */
	if ((fd = socket(sock_af(addr), addr->type, PF_UNSPEC)) == -1)
		return (NULL);
	(void) fcntl(fd, F_SETFD, 1);
	if (setjmp(timejump)) {
		closeSys(fd);
		/* restore old mask (saved below, which executes earlier
		 * due to setjmp().
		 */
		(void) sigprocmask(SIG_SETMASK, &oset, NULL);
		/* XXX - this can cause late/missed itimers in appl. */
		(void) setitimer(ITIMER_REAL, &oldtimeout, NULL);
		errno = ETIMEDOUT;
		return (NULL);
	}
	timerclear(&timeout.it_interval);
	timeout.it_value.tv_sec = 15;	/* dramatically short, isn't it? */
	timeout.it_value.tv_usec = 0;	/* (has to be shorter than appl.'s) */
	if (setitimer(ITIMER_REAL, &timeout, &oldtimeout) == -1) {
		int save;
 err:
		save = errno;
		closeSys(fd);
		errno = save;
		return (NULL);
	}
	/* Block interrupts whose handlers often longjmp(). */
	sigemptyset(&set);
	sigaddset(&set, SIGINT);
	if (sigprocmask(SIG_BLOCK, &set, &oset) == -1) {
		int save;
 err2:
		save = errno;
		(void) setitimer(ITIMER_REAL, &oldtimeout, NULL);
		errno = save;
		goto err;
	}
	oldalrm = signal(SIGALRM, sigalrm);
	if (connectSys(fd, &addr->addr, sizeof addr->addr) == -1)
		goto err2;
	(void) signal(SIGALRM, oldalrm);
	/* XXX - this can cause late/missed itimers in appl. */
	setitimer(ITIMER_REAL, &oldtimeout, NULL);
	(void) sigprocmask(SIG_SETMASK, &oset, NULL);
	if (!(conn = fdopen(fd, "r+"))) {
		int save = errno;
		closeSys(fd);
		errno = save;
		return (NULL);
	}
	setvbuf(conn, (char *)NULL, _IOLBF, 0);

	/*
	 * Parse initial banner from server.
	 */
	while (!proxsrvResp(conn, buf, sizeof buf, &code))
		NULL;
	if ((code & 0700) != 0200) {
		fclose(conn);
		syslog(LOG_WARNING, "proxsrvOpen(%s): \"%s\"",
		       sock_ntoa(addr), buf);
		errno = ECONNRESET;
		return (NULL);
	}

	/*
	 * Connection is up.  Cache and return it.
	 */
	if (!(psp = ZALLOC(struct proxSrv, ZAL_ZERO))) {
		int save = errno;
		fclose(conn);
		errno = save;
		return (NULL);
	}
	psp->addr = *addr;
	psp->fp = conn;
	ListAdd(&State.srv, psp);
	return (conn);
}

void
proxsrvClose(FILE *fp) {
	struct proxSrv *psp, *p, *n;

	fclose(fp);
	p = NULL;
	ListEach2(&State.srv, psp, n) {
		if (psp->fp == fp) {
			ListUnlink(&State.srv, psp, p);
			free(psp);
			break;
		}
		p = psp;
	}
}

void
proxsrvCloseAll() {
	struct proxSrv *psp, *p, *n;

	p = NULL;
	ListEach2(&State.srv, psp, n) {
		fclose(psp->fp);
		ListUnlink(&State.srv, psp, p);
		free(psp);
	}
}

int
proxsrvResp(FILE *fp, char *buf, int buflen, int *code) {
	int x;

	proxsrvInit();
	if (ferror(fp)) {
		proxyTr(PROXTRC_PROTO, "!!! Previous error\n");
		*code = 0577;
		return (-1);
	}
	if (!fgets(buf, buflen, fp)) {
		proxyTr(PROXTRC_PROTO, "!!! Unexpected %s\n",
			ferror(fp) ? "error" : "EOF");
		*code = 0577;
		return (-1);
	}
	x = strlen(buf);
	while (x && strchr("\r\n", buf[x-1]))
		buf[--x] = '\0';
	x = arpa_parsecode(buf, code);
	proxyTr(PROXTRC_PROTO, "<<< %s  (%03o, %d)\n", buf, *code, x);
	return (x);
}
