/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* sys_listen.c - "listen" system call replacement for libproxynet
 * vix 30mar94 [cloned from sys_connect [02feb94 [original]]]
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: sys_listen.c,v 1.8 1996/05/16 04:48:39 vixie Exp $";
#endif

#include "vixutil.h"

#include <sys/syscall.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>

#include "defs.h"
#include "messages.h"


int
listenSys(int s, int backlog) {
	if (sock_fdtype(s) != -1)
		proxyTr(PROXTRC_SYSCALL, "listenSys(%d, %d)\n", s, backlog);
	return (syscall(SYS_listen, s, backlog));
}


int
listenProxy(const struct sock *cla, const struct sock *spa, void *pcb, int s) {
	static const char funcName[] = "listenProxy";
	char buf[ARPA_MAXSTR], tmp[MEDIUM];
	FILE *resp;
	int code;

	/* If there are no proxy servers, we have nothing to do here. */
	if (!proxyProxsrvs_p(pcb))
		return (0);

	strncpy(tmp, sock_ntoa(spa), sizeof tmp);
	sprintf(buf, "lstn %s %s", sock_ntoa(cla), tmp);
	if (!(resp = proxyTransaction(pcb, buf))) {
		syslog(LOG_INFO, proxmsg_CantConn,
		       funcName, sock_ntoa(cla));
		errno = ENETUNREACH;
		return (-1);
	}
	while (!proxsrvResp(resp, buf, sizeof buf, &code))
		NULL;
	if ((code & 0700) != 0200) {	/* XXX - need to check entire code? */
		syslog(LOG_INFO, proxmsg_RespNotOK,
		       funcName, sock_ntoa(cla), buf);
		errno = ENETUNREACH;
		return (-1);
	}
	proxfdSetT(s, fdTlisten);
	proxfdSetL(s, pcb, cla, spa);
	return (0);
}


struct eachUab {
	struct sock	*cla;
	int		wildport;
	int		s;
};


static int
eachPcb(void *pcb, int t, void *uap) {
	const struct eachUab *uab = (struct eachUab *)uap;
	struct sock spa, cla;
	struct sockaddr sa;
	int len;

	/* We only handle IP streams right now. (XXX) */
	if (!sock_isstream(uab->cla) || sock_af(uab->cla) != AF_INET)
		goto done;

	/* Get the real sockaddr, so we can check to see if it has a proxy. */
	len = sizeof(struct sockaddr);
	if (getsocknameSys(t, &sa, &len) == -1)
		goto done;
	cla = sock_sock(sock_fdtype(t), &sa);

	/* Build temporary sock for service proxy address. */
	spa = *uab->cla;
	((struct sockaddr_in *)&spa.addr)->sin_addr.s_addr =
		(getsocknameProxy(t, &cla, pcb) == 0)
		/* XXX - non-ftp apps might hate this a LOT. */
		? ((struct sockaddr_in *)&cla.addr)->sin_addr.s_addr
		: INADDR_ANY;	/* XXX - should use specific local address? */
	if (uab->wildport)
		((struct sockaddr_in *)&spa.addr)->sin_port = htons(0);

	/* Make it happen. */
	(void) listenProxy(uab->cla, &spa, pcb, uab->s);

 done:
	return (0);
}


/*
 * We will set up a proxy listener if either:
 *	The local service address matches a proxy entry's list, OR:
 *	An associated socket is itself proxied.
 */


int
listen(int s, int backlog) {
	struct sockaddr sa;
	struct sock cla;
	int save, len, t;
	struct eachUab uab;

	if (listenSys(s, backlog) == -1)
		return (-1);
	save = errno;

	len = sizeof(struct sockaddr);
	if (getsocknameSys(s, &sa, &len) == -1)   /* sock_name() would proxy */
		goto done;
	cla = sock_sock(sock_fdtype(s), &sa);

	if (proxyEachSvcMatch(&cla, listenProxy, s) != 0)
		goto done;

	/* No luck so far.  Travel up the association chain. */
	uab.cla = &cla;
	uab.wildport = (proxfdGetW(s) > 0);
	uab.s = s;
	t = s;
	while ((t = proxfdGet(t)) != -1)
		(void) proxfdEachP(t, eachPcb, &uab);
 done:
	errno = save;
	return (0);
}
