/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* fgettok.c - get next token from file
 * vixie 17feb94 [original]
 */

#if !defined(LINT) && !defined(CODECENTER)
static char rcsid[] = "$Id: fgettok.c,v 1.3 1996/05/16 04:48:45 vixie Exp $";
#endif

#include "vixutil.h"
#include <stdio.h>

#define TIME 0
#define SPACE 1
#define OPTIMIZE SPACE

#if (OPTIMIZE == SPACE)
# define nextch() (ch = fgetc(f))
#else
# define nextch() (ch = getc(f))
#endif

/* fgettok()
 *	skip leading white space, find token using given separator set.
 *	skip to-end-of-line comments automatically.
 * returns:
 *	-1 if EOF encountered before useful data
 *	0 if the token plus a \0 would be too long for the buffer
 *	1..255 if something useful was found; this is the token separator found
 */
int
fgettok(FILE *f, char *buf, int buflen,
	const char *sepset, const char *comset
	) {
	register int ch;
	register int sepspace = FALSE;
	register char *bufp = buf;
	register char *endp = buf + buflen - 1;
	register const char *sepp = sepset;

	/*
	 * Determine whether there are blanks in the sepset; if so,
	 * arrange to alias them later (any blank in the input will
	 * match it).  Newlines are made exempt from this aliasing.
	 */
	while ((ch = *sepp++) != '\0')
		if (isspace(ch) && ch != '\n') {
			sepspace = TRUE;
			break;
		}

	/*
	 * Skip leading blanks, and if a comset was specified, skip
	 * comments and blanks that follow comments, until no comment
	 * is found.
	 */
	do {nextch();} while (ch != EOF && isspace(ch));
	if (comset)
		while (strchr(comset, (char)ch)) {
			do {nextch();} while (ch != EOF && ch != '\n');
			do {nextch();} while (ch != EOF && isspace(ch));
		}

	/*
	 * Move characters across until we either run out, hit a separator,
	 * or hit an aliased blank from the separator.  Filling the output
	 * buffer is an error exit with the buffer undefined (unterminated).
	 */
	while (ch != EOF &&
	       !strchr(sepset, (char)ch) &&
	       (!sepspace || ch == '\n' || !isspace(ch))
	       ) {
		*bufp++ = (char) ch;
		if (bufp >= endp)
			return (0);
		nextch();
	}

	/* Did we hit EOF before any output characters could be copied? */
	if (ch == EOF && bufp == buf)
		return (-1);

	/* Terminate the buffer and exit normally. */
	*bufp = '\0';
	return (ch);
}
