/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* connect.c - proxynet connect support
 * vix 05may94 [split out of session.c]
 */

#if !defined(lint) && !defined(codecenter)
static char rcsid[] = "$Id: connect.c,v 1.2 1996/05/16 04:48:56 vixie Exp $";
#endif

#include "externs.h"
#include "defs.h"

static jmp_buf	jumpalarm;
static char	msgBuf[ARPA_MAXSTR + 50];

static void
sigalrm(int junk) {
	longjmp(jumpalarm, 1);
}

int
connStream(const struct sock *dest) {
	volatile int cli, status;
	int srv, rem, pid, size, save;
	struct prox_tup new;
	struct prox_hnd *dbe;
	struct sock sess;
	fd_set fds;
	int on = 1;

	memset(&new, 0, sizeof new);

	/* client socket will be in the same domain as the session */
	if (sock_name(STDIN_FILENO, &sess) == -1) {
#if notyet
		return (eemit("getsockname(STDIN)", errno));
#else
		memset(&sess, 0, sizeof sess);
		sess.addr.sa_family = AF_INET;
		sess.addr.sa_len = sizeof(struct sockaddr_in);
		((struct sockaddr_in *)&sess.addr)->sin_addr.s_addr =
			htonl(0x7f000001);
		((struct sockaddr_in *)&sess.addr)->sin_port =
			IPPORT_USERRESERVED + 1;
		sess.type = SOCK_STREAM;
#endif
	}
	if (sock_af(&sess) != AF_INET) {
		return (emit(0541, "non-internet streams not supported\n"));
	}
	if (sock_peer(STDIN_FILENO, &new.ctl) == -1) {
#ifdef notyet
		return (eemit("getpeername(STDIN)", errno));
#else
		new.ctl = sess;
		((struct sockaddr_in *)&new.ctl.addr)->sin_port =
			IPPORT_USERRESERVED;
#endif
	}

	/* get our server socket set up, for client<->proxy connection */
	srv = socket(sock_af(&sess), sess.type, PF_UNSPEC);
	if (srv == -1) {
		return (eemit("client socket", errno));
	}
	new.cpa = sess;
	((struct sockaddr_in *)&new.cpa.addr)->sin_port = 0;
	if (sock_bind(srv, &new.cpa, SOCK_F_RSVD) == -1) {
		eclose(srv);
		return (eemita("bind/CPA", errno, &new.cpa));
	}
	/* now get actual name, with port filled in by bind() */
	if (sock_name(srv, &new.cpa) == -1) {
		eclose(srv);
		return (eemit("getsockname(CPA)", errno));
	}
	if (listen(srv, SOMAXCONN) == -1) {
		eclose(srv);
		return (eemit("listen(CPA)", errno));
	}

	/* remote socket will be in the domain specified by command. */
	rem = socket(sock_af(dest), dest->type, PF_UNSPEC);
	if (rem == -1) {
		eclose(srv);
		return (eemit("remote socket", errno));
	}
	(void) setsockopt(rem, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on);
	/* fire off our remote connection, then get both names */
	if (setjmp(jumpalarm) > 0) {
		syslog(LOG_INFO, "connect: timeout");
		eclose(srv); eclose(rem);
		return (emit(0523, "remote connection timed out\n"));
	}
	signal(SIGALRM, sigalrm);
	alarm(15);
	do {
		status = connect(rem, &dest->addr, sizeof(struct sockaddr));
	} while (status == -1 && errno == EINTR);
	save = errno;
	alarm(0);
	signal(SIGALRM, SIG_DFL);
	errno = save;
	if (status == -1 && errno == EISCONN) {
		syslog(LOG_DEBUG, "(harmless?) connect: %m");
		status = 0;
	}
	if (status == -1) {
		eclose(srv); eclose(rem);
		return (eemit("remote connect", errno));
	}
	if (sock_name(rem, &new.spa) == -1) {
		eclose(srv); eclose(rem);
		return (eemit("getsockname(SPA)", errno));
	}
	if (sock_peer(rem, &new.sra) == -1) {
		eclose(srv); eclose(rem);
		return (eemit("getsockname(SRA)", errno));
	}

	/* fork.  parent will send protocol ACK and return to command loop;
	 * child will continue below.
	 */
	switch (pid = fork()) {
	case -1:
		/* error */
		eclose(srv); eclose(rem);
		return (eemit("fork", errno));
	case 0:
		/* child */
		break;
	default:
		/* parent */
		close(srv); close(rem);
		sprintf(msgBuf, "<%s> listening (pid %d)\n",
			sock_ntoa(&new.cpa), pid);
		return (emit(0201, msgBuf));
	}

	/* now running in child.  close session descriptors. */
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
	/* wait for connection, then get peer's name */
	if (setjmp(jumpalarm) > 0) {
		syslog(LOG_INFO, "accept: timeout");
		myExit(EX_PROTOCOL);
	}
	signal(SIGALRM, sigalrm);
	alarm(15);
	size = sizeof new.cla.addr;
	do {
		cli = accept(srv, &new.cla.addr, &size);
	} while (cli == -1 && errno == EINTR);
	save = errno;
	alarm(0);
	signal(SIGALRM, SIG_DFL);
	errno = save;
	if (cli == -1) {
		syslog(LOG_INFO, "accept: %m");
		myExit(EX_OSERR);
	}
	new.cla.type = sess.type;
	/* we are done with the server socket */
	close(srv);

	/* make sure the connection came from the same host as the session */
	if (((struct sockaddr_in *)&new.cla.addr)->sin_addr.s_addr !=
	    ((struct sockaddr_in *)&new.ctl.addr)->sin_addr.s_addr) {
		strcpy(msgBuf, sock_ntoa(&new.cla));
		syslog(LOG_WARNING, "address mismatch (cla %s) != (ctl %s)",
		       msgBuf, sock_ntoa(&new.ctl));
		myExit(EX_PROTOCOL);
	}

	/* we know all four addresses now; enter into connection database. */
	if (!(dbe = dbEnter(&new)))
		myExit(EX_OSERR);

	strcpy(msgBuf, sock_ntoa(&new.cla));
	syslog(LOG_INFO, "connected %s -> %s", msgBuf, sock_ntoa(&new.sra));

#if bad_idea
	FD_ZERO(&fds);
	FD_SET(cli, &fds);
	FD_SET(rem, &fds);
	FD_SET(fileno(dbe->fp), &fds);
	/* XXX:	we're going to close syslog's file descriptors, too. */
	closeFdSet(&fds, 0);
#endif

	if (copy2way_exec(cli, rem, 8*1024) == -1)
		syslog(LOG_NOTICE, "copy2way_exec: %m");
	close(cli);  close(rem);

	strcpy(msgBuf, sock_ntoa(&new.cla));
	syslog(LOG_INFO, "shutdown %s -> %s", msgBuf, sock_ntoa(&new.sra));

	dbRemove(dbe);
	myExit(EX_OK);
}
