/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* database.c - database for proxytcpd
 * vix 01apr94 [added service view]
 * vix 24feb94 [original]
 */

#if !defined(lint) && !defined(codecenter)
static char rcsid[] = "$Id: database.c,v 1.4 1996/05/16 04:48:56 vixie Exp $";
#endif

#define _DATABASE_C

#include "externs.h"
#include "defs.h"

static char dbDir[] = PATH_DBDIR;
#ifdef XXX
static char svcDir[] = PATH_SVCDIR;
#endif


static int
dbValidName(const char *name) {
	int i;

	if (strlen(name) != 14)
		return (FALSE);
	for (i = 0;  i < 10;  i++)
		if (!isdigit(name[i]))
			return (FALSE);
	if (name[10] != '.')
		return (FALSE);
	for (i = 11;  i < 14;  i++)
		if (!isdigit(name[i]))
			return (FALSE);
	return (TRUE);
}


void
dbInit() {
	if (mkdir(dbDir, DIRMODE) == -1 && errno != EEXIST) {
		syslog(LOG_ERR, "mkdir(%s): %m", dbDir);
		myExit(EX_UNAVAILABLE);
	}
#ifdef XXX
	if (mkdir(svcDir, DIRMODE) == -1 && errno != EEXIST) {
		syslog(LOG_ERR, "mkdir(%s): %m", svcDir);
		myExit(EX_UNAVAILABLE);
	}
#endif
}


#ifdef XXX
int
dbSvcEnter(const struct sock *svc) {
	char buf[MAXPATHLEN], namebuf[MAXPATHLEN], linkname[SMALL];
	char *sp = sock_ntoa(svc), *dp = buf, ch, *tmp;
	int len;

	sprintf(linkname, "%ld", getpid());
	if (!dbValidName(linkname) ||
	    (sizeof(dbDir) + strlen(linkname) + 5 > MAXPATHLEN)) {
		errno = ENAMETOOLONG;
		return (-1);
	}
	sprintf(namebuf, "%s/%s", dbDir, linkname);

	/* / -> |
	 * | -> ||
	 */
	while (ch = *sp++) {
		if (ch == '|')
			*dp++ = '|';
		if (ch == '/')
			*dp++ = '|';
		else
			*dp++ = ch;
		if (dp >= &buf[sizeof buf - 1]) {
			syslog(LOG_ERR, "dbEnterSvc(%s): too long",
			       sock_ntoa(svc));
			errno = ENAMETOOLONG;
			return (-1);
		}
	}
	*dp = '\0';

	tmp = malloc(sizeof PATH_SVCDIR + (dp - buf) + strlen(linkname) + 42);
	len = sprintf(tmp, "%s/%s", svcDir, buf);
	if (mkdir(tmp, DIRMODE) == -1 && errno != EEXIST) {
		int save = errno;

		syslog(LOG_ERR, "mkdir(%s): %m", tmp);
		free(tmp);
		errno = save;
		return (-1);
	}
	(void) sprintf(tmp+len, "/%s", linkname);

	if (link(namebuf, tmp) == -1) {
		int save = errno;

		syslog(LOG_ERR, "link(%s, %s): %m", namebuf, tmp);
		free(tmp);
		errno = save;
		return (-1);
	}

	free(tmp);
	return (0);
}
#endif


void
dbForeach(dbUar uar, void *ucp) {
	DIR *dbd;
	struct dirent *dp;

	dbd = opendir(dbDir);
	if (!dbd) {
		syslog(LOG_WARNING, "opendir(%s): %m", dbDir);
		return;
	}
	while (dp = readdir(dbd)) {
		if (!dbValidName(dp->d_name))
			continue;
		if ((*uar)(dp->d_name, ucp) == 1)
			break;
	}
	closedir(dbd);
}

/* an entry is stale if it is more than a few seconds old and we can lock it.
 *
 * return value:
 *	 <0 = invalid filename, no such entry, too new, or stale (deleted)
 *	>=0 = entry is not stale, retval is open(readable) filedes
 */
int
dbOpen(const char *filename) {
	char namebuf[MAXPATHLEN];
	struct stat statbuf;
	time_t now;
	int fd;

	if (!dbValidName(filename) ||
	    (sizeof(dbDir) + strlen(filename) + 5 > MAXPATHLEN)) {
		errno = ENAMETOOLONG;
		return (-1);
	}
	sprintf(namebuf, "%s/%s", dbDir, filename);
	if ((fd = open(namebuf, O_RDONLY)) < 0 || fstat(fd, &statbuf) < 0)
		return (-1);
	time(&now);
	if (now - statbuf.st_mtime < 2)
		sleep(1);
	if (flock(fd, LOCK_EX|LOCK_NB) == 0) {
		syslog(LOG_DEBUG, "removing stale entry \"%s\"", namebuf);
		remove(namebuf);	/* POSIX for "unlink" */
		close(fd);		/* implicitly un-flocks */
		errno = ETIMEDOUT;
		return (-1);
	}
	return (fd);
}

#define okCTL	0x01
#define okCLA	0x02
#define	okCPA	0x04
#define	okSPA	0x08
#define	okSRA	0x10
#define	okFlags	0x20
#define	okAll	(okCTL|okCLA|okCPA|okSPA|okSRA|okFlags)

/* return the address of a static tuple read from file, or NULL
 */
struct prox_tup *
dbFetch(const char *filename) {
	int fd, ok;
	FILE *fp;
	char tok1[MEDIUM], tok2[MEDIUM], *msg;
	static struct prox_tup ret;

	bzero((char *)&ret, sizeof ret);
	if ((fd = dbOpen(filename)) < 0)
		return (NULL);
	if (!(fp = fdopen(fd, "r"))) {
		int save = errno;
		close(fd);
		errno = save;
		return (NULL);
	}
	ok = 0;
	msg = NULL;
	while (0 < fgettok(fp, tok1, sizeof tok1, " ", NULL) &&
	       0 < fgettok(fp, tok2, sizeof tok2, "\n", NULL)) {
		if (!strcmp(tok1, "ctl")) {
			if (!sock_aton(tok2, &ret.ctl)) {
				msg = "bad ctl";
				break;
			}
			ok |= okCTL;
		} else if (!strcmp(tok1, "cla")) {
			if (!sock_aton(tok2, &ret.cla)) {
				msg = "bad cla";
				break;
			}
			ok |= okCLA;
		} else if (!strcmp(tok1, "cpa")) {
			if (!sock_aton(tok2, &ret.cpa)) {
				msg = "bad cpa";
				break;
			}
			ok |= okCPA;
		} else if (!strcmp(tok1, "spa")) {
			if (!sock_aton(tok2, &ret.spa)) {
				msg = "bad spa";
				break;
			}
			ok |= okSPA;
		} else if (!strcmp(tok1, "sra")) {
			if (!sock_aton(tok2, &ret.sra)) {
				msg = "bad sra";
				break;
			}
			ok |= okSRA;
		} else if (!strcmp(tok1, "flg")) {
			if (tok2[0] != '0' ||
			    (tok2[1] != 'x' && tok2[1] != 'X') ||
			    !str_ishexnum(tok2)) {
				msg = "bad flg";
				break;
			}
			ret.flags = strtol(tok2, NULL, 16);
			ok |= okFlags;
		} else {
			msg = "bad token";
			break;
		}
	}
	if (ok != okAll && !msg)
		msg = "missing datum";
	if (msg) {
		syslog(LOG_WARNING, "\"%s/%s\": %s", dbDir, filename, msg);
		fclose(fp);
		errno = EX_SOFTWARE;
		return (NULL);
	}
	fclose(fp);
	return (&ret);
}

/* this returns an flock()'d handle; fclose() it to release the lock.
 *
 * NULL indicates an error; check errno.
 */
struct prox_hnd *
dbEnter(const struct prox_tup *new) {
	char namebuf[MAXPATHLEN];
	struct prox_hnd *ret;
	pid_t pid = getpid();
	static int disturb;
	int i, fd;
	FILE *fp;

	for (i = 0;  i < 10;  i++) {
		sprintf(namebuf, "%010ld.%03d", (long) pid, disturb);
		if ((fd = dbOpen(namebuf)) < 0)
			break;
		if (++disturb > 999)
			disturb = 0;
	}
	if (fd >= 0) {
		/* file exists and we couldn't lock it.  complain and fail. */
		close(fd);
		syslog(LOG_WARNING, "duplicate entry \"%s/%s\"",
		       dbDir, namebuf);
		return (NULL);
	}
	sprintf(namebuf, "%s/%010ld.%03d", dbDir, (long) pid, disturb);
	if ((fd = open(namebuf, O_WRONLY|O_CREAT|O_EXCL, 0660)) < 0) {
		syslog(LOG_WARNING, "can't create \"%s\": %m", namebuf);
		return (NULL);
	}
	(void) fcntl(fd, F_SETFD, 1);
	if (!(fp = fdopen(fd, "w"))) {
		syslog(LOG_WARNING, "can't fdopen \"%s\": %m", namebuf);
		close(fd);
		return (NULL);
	}
	if (flock(fd, LOCK_EX|LOCK_NB) < 0) {
		syslog(LOG_WARNING, "can't flock \"%s\": %m", namebuf);
		fclose(fp);
		return (NULL);
	}
	fprintf(fp, "ctl %s\n", sock_ntoa(&new->ctl));
	fprintf(fp, "cla %s\n", sock_ntoa(&new->cla));
	fprintf(fp, "cpa %s\n", sock_ntoa(&new->cpa));
	fprintf(fp, "spa %s\n", sock_ntoa(&new->spa));
	fprintf(fp, "sra %s\n", sock_ntoa(&new->sra));
	fprintf(fp, "flg 0x%x\n", new->flags);
	if (fflush(fp) < 0) {
		syslog(LOG_WARNING, "can't fflush \"%s\": %m", namebuf);
		fclose(fp);
		return (NULL);
	}
	if (!(ret = malloc(sizeof *ret))) {
		syslog(LOG_WARNING, "malloc failed for \"%s\": %m", namebuf);
		fclose(fp);
		return (NULL);
	}
	memset(ret, 0, sizeof *ret);
	ret->fp = fp;
	ret->disturb = disturb;
	ret->pid = pid;
	return (ret);
}

void
dbRemove(struct prox_hnd *handle) {
	char namebuf[MAXPATHLEN];
	struct stat sb, fsb;

	if (!handle)
		return;

	sprintf(namebuf, "%s/%010ld.%03d",
		dbDir, (long)handle->pid, handle->disturb);
	if (stat(namebuf, &sb) < 0) {
		syslog(LOG_WARNING, "stat(%s): %m", namebuf);
		return;
	}
	if (fstat(fileno(handle->fp), &fsb) < 0) {
		syslog(LOG_WARNING, "fstat(%d): %m", fileno(handle->fp));
		return;
	}
	if (sb.st_dev != fsb.st_dev || sb.st_ino != fsb.st_ino) {
		syslog(LOG_WARNING,
		       "file %s changed: path(0x%lx,%ld) != file(0x%lx,%ld)",
		       sb.st_dev, sb.st_ino, fsb.st_dev, fsb.st_ino);
		return;
	}
	remove(namebuf);
	fclose(handle->fp);	/* does an unlock */
	free(handle);
}

void
dbClose(struct prox_hnd *handle) {
	if (!handle)
		return;
	fclose(handle->fp);     /* does an unlock */
	free(handle);
}


int
dbMtime(time_t *mtime) {
	struct stat sb;

	if (stat(dbDir, &sb) < 0)
		return (-1);
	*mtime = sb.st_mtime;
	return (0);
}
