/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* listen.c - proxynet listener support
 * vix 05may94 [split out of session.c]
 */

#if !defined(lint) && !defined(codecenter)
static char rcsid[] = "$Id: listen.c,v 1.3 1996/05/16 04:48:56 vixie Exp $";
#endif

#include "externs.h"
#include "defs.h"

struct listener {
	ListLink(struct listener)
	struct sock	cla;
	int		flags;
#define	LSTN_WEIGHT	0x0001
#define	LSTN_MARKED	0x0002
};

struct lstnCtx {
	const struct sock	*spa;
	struct listener		*best_lp;
	int			best_weight;
};

static ListHeadV(Listeners, struct listener);
static char	msgBuf[ARPA_MAXSTR + 50];

static int	listener(int srv, struct sock s_spa),
		listener_child(int so_sra,
			       const struct sock *s_spa,
			       const struct sock *s_sra,
			       const struct sock *s_cla);

int
lstnStream(const struct sock *s_cla, struct sock *s_spa) {
	struct prox_tup new;
	struct prox_hnd *dbe;
	struct lstnDbe	*ldbe;
	int srv, pid, code, n, save;
	const int m = 8, t = 30, on = 1;

	/* get our server socket set up, for proxy<->client service */
	srv = socket(sock_af(s_spa), s_spa->type, PF_UNSPEC);
	if (srv == -1)
		return (eemit("client socket", errno));

	if (setsockopt(srv, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on) == -1)
		return (eemit("SO_REUSEADDR", errno));

	/* Try bind() several times since old servers can linger in TIME_WAIT
	 * after they die.  This can take a while, and unfortunately we will
	 * delay our client while we check this out...
	 */
	for (n = 1;  n <= m;  n++) {
		if (bind(srv, &s_spa->addr, sizeof(struct sockaddr)) == 0) {
			errno = -1;
			break;
		}
		save = errno;
		syslog(LOG_NOTICE, "bind(%s) (try %d/%d): %m",
		       sock_ntoa(s_spa), n, m);
		if (errno == EADDRINUSE) {
			if (n < m) {
				emit1(0231, '-', "EADDRINUSE, sleeping", 0);
				sleep(t);
			}
		} else {
			n = m;
		}
		errno = save;
	}
	if (errno != -1)
		return (eemita("bind", errno, s_spa));
	if (sock_name(srv, s_spa) == -1)
		return (eemit("sock_name", errno));

	memset(&new, 0, sizeof new);
	if (sock_peer(STDIN_FILENO, &new.ctl) == -1)
		return (eemit("getpeername(STDIN)", errno));
	new.cla = *s_cla;
	new.spa = *s_spa;
	new.cpa = *s_spa;	/* placeholder */
	new.sra = *s_spa;	/* placeholder */
	new.flags |= PT_LISTENER | PT_CLIENT;
	/* XXX:	we should not enter this into the database if a duplicate
	 *	(using {CLA,SPA} as key) already exists.
	 */
	if (!(dbe = dbEnter(&new)))
		return (emit(0544, "can't enter dbe - see syslog\n"));
	strcpy(msgBuf, sock_ntoa(s_spa));
	syslog(LOG_INFO, "listening %s <- %s", sock_ntoa(s_cla), msgBuf);

	/* fork.  parent will send protocol ACK and return to command loop;
	 * child will continue below.
	 */
	switch (pid = fork()) {
	case -1:
		save = errno;
		dbRemove(dbe);
		errno = save;
		return (eemit("fork", errno));
	case 0:
		/* child */
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
		dbClose(dbe);
		(void) setsid();
		myExit(listener(srv, *s_spa));
	default:
		/* parent */
		close(srv);
		code = 0201;
		sprintf(msgBuf, "<%s> listening (pid %d)\n",
			sock_ntoa(s_spa), pid);

		/* put this dbe on a list so we can clean it up on exit. */
		ldbe = malloc(sizeof *ldbe);
		if (!ldbe)
			return (eemit("malloc", errno));
		memset(ldbe, 0, sizeof *ldbe);
		ldbe->dbe = dbe;
		ListAdd(&lstnDbeL, ldbe);

		return (emit(code, msgBuf));
	}
	/*NOTREACHED*/
}

static int
lstnEach(const char *key, void *ucp) {
	struct lstnCtx *ctx = (struct lstnCtx *)ucp;
	struct prox_tup *ent = dbFetch(key);
	struct listener *lp, *found_lp;
	int weight;

	if (!ent) {
		syslog(LOG_DEBUG, "cannot fetch %s: %m", key);
		return (0);	/* "keep going" */
	}
	if (!sock_same(ctx->spa, &ent->spa)) {
		return (0);	/* "keep going" */
	}

	found_lp = NULL;
	ListEach(&Listeners, lp) {
		if (sock_same(&lp->cla, &ent->cla)) {
			found_lp = lp;
			break;
		}
		else {
			strcpy(msgBuf, sock_ntoa(&ent->cla));
			syslog(LOG_DEBUG, "!sock_same([%d]%s, [%d]%s)",
			       lp->cla.addr.sa_len, sock_ntoa(&lp->cla),
			       ent->cla.addr.sa_len, msgBuf);
		}
	}
	if (found_lp) {
		lp = found_lp;
	} else {
		lp = malloc(sizeof *lp);
		if (!lp) {
			syslog(LOG_WARNING, "malloc failed: %m");
			return (1);	/* "stop here" */
		}
		memset(lp, 0, sizeof *lp);
		lp->cla = ent->cla;
		ListAdd(&Listeners, lp);
		syslog(LOG_DEBUG, "new CLA: %s", sock_ntoa(&ent->cla));
		/* XXX:	should this new one have to wait its turn
		 *	or should it get the very next round robin?
		 */
	}
	lp->flags |= LSTN_MARKED;
	weight = (lp->flags & LSTN_WEIGHT) ? 1 : 0;
	if (weight < ctx->best_weight) {
		ctx->best_weight = weight;
		ctx->best_lp = lp;
	}
	return (0);	/* "keep going" */
}

static struct sock *
listenerSelect(const struct sock *s_spa) {
	struct lstnCtx	ctx;
	struct listener	*lp, *p, *n;

	/*
	 * Implement round-robin while picking up new listeners.
	 */
	ListEach(&Listeners, lp)
		lp->flags &= ~LSTN_MARKED;

	memset(&ctx, 0, sizeof ctx);
	ctx.spa = s_spa;
	ctx.best_weight = INT_MAX;
	dbForeach(lstnEach, &ctx);
	/* Check for zombie client listeners. */
	p = NULL;
	ListEach2(&Listeners, lp, n) {
		if (!(lp->flags & LSTN_MARKED)) {
			syslog(LOG_INFO, "client listener <%s> has gone away",
			       sock_ntoa(&lp->cla));
			ListUnlink(&Listeners, lp, p);
			free(lp);
		} else {
			p = lp;
		}
	}
	if (ctx.best_weight != 0) {
		/* Everybody has had a turn. */
		ListEach(&Listeners, lp)
			lp->flags &= ~LSTN_WEIGHT;
	}
	if (!ctx.best_lp)
		return (NULL);
	ctx.best_lp->flags |= LSTN_WEIGHT;
	return (&ctx.best_lp->cla);
}

static int
listener(int srv, struct sock s_spa) {
	struct prox_tup new;

	if (listen(srv, SOMAXCONN) == -1) {
		syslog(LOG_ERR, "listen: %m");
		return (EX_OSERR);
	}
	syslog(LOG_INFO, "listening %s", sock_ntoa(&s_spa));

	/*
	 * Looping, collecting new connections.  We'll block in accept().
	 */
	for (;;) {
		struct sock s_sra, *s_cla;
		struct sockaddr cli_sa;
		int pid, cli, size;

		size = sizeof cli_sa;
		do {
			cli = accept(srv, &cli_sa, &size);
		} while (cli == -1 && errno == EINTR);
		if (cli == -1) {
			syslog(LOG_ERR, "accept: %m");
			return (EX_OSERR);
		}
		s_sra.type = sock_fdtype(cli);
		s_sra.addr = cli_sa;

		s_cla = listenerSelect(&s_spa);
		if (!s_cla) {
			syslog(LOG_INFO,
			       "refusing conn from <%s>; no clients found",
			       sock_ntoa(&s_sra));
			break;
		}

		/*
		 * Each connection runs in its own subprocess; that way we
		 * can continue looping here and waiting for more cnxns.
		 */
		switch (pid = fork()) {
		case -1:
			syslog(LOG_ERR, "fork: %m");
			return (EX_OSERR);
		case 0:
			/* child */
			close(srv);
			switch (listener_child(cli, &s_spa, &s_sra, s_cla)) {
			case -1:
				myExit(errno);
				/*NOTREACHED*/
			default:
				myExit(EX_OK);
				/*NOTREACHED*/
			}
			/*NOTREACHED*/
		default:
			/* parent */
			close(cli);
			break;
		}
	}
}

static int
listener_child(int so_sra,
	       const struct sock *s_spa,
	       const struct sock *s_sra,
	       const struct sock *s_cla) {
	struct prox_tup	new;
	struct prox_hnd	*dbe;
	int		so_cla;
	fd_set		fds;

	memset(&new, 0, sizeof new);
	new.ctl = *s_spa;	/* placeholder */
	new.spa = *s_spa;
	new.sra = *s_sra;

	/* client listener socket will be in domain specified by command. */
	so_cla = socket(sock_af(s_cla), s_cla->type, PF_UNSPEC);
	if (so_cla == -1) {
		syslog(LOG_WARNING, "socket: %m");
		return (-1);
	}
	utime(PATH_DBDIR, NULL);	/* XXX ugly */
	/* fire off client listener connection, then get both names */
	/* XXX:	that this needs a timeout */
	if (connect(so_cla, &s_cla->addr, sizeof(struct sockaddr)) == -1) {
		syslog(LOG_WARNING, "connect(%s): %m", sock_ntoa(s_cla));
		return (-1);
	}
	if (sock_name(so_cla, &new.cpa) == -1) {
		syslog(LOG_WARNING, "sock_name: %m");
		return (-1);
	}
	if (sock_peer(so_cla, &new.cla) == -1) {
		syslog(LOG_WARNING, "sock_peer: %m");
		return (-1);
	}

	if (!(dbe = dbEnter(&new))) {
		errno = EX_UNAVAILABLE;
		return (-1);
	}

	strcpy(msgBuf, sock_ntoa(&new.cla));
	syslog(LOG_INFO, "connected %s <- %s", msgBuf, sock_ntoa(&new.sra));

	if (copy2way_exec(so_sra, so_cla, 8*1024) == -1)
		syslog(LOG_NOTICE, "copy2way_exec: %m");
	eclose(so_sra);  eclose(so_cla);

	strcpy(msgBuf, sock_ntoa(&new.cla));
	syslog(LOG_INFO, "shutdown %s <- %s", msgBuf, sock_ntoa(&new.sra));

	dbRemove(dbe);
	return (0);
}
