/*
 * Copyright (c) 1994-1996 by Vixie Enterprises.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND VIXIE ENTERPRISES DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL VIXIE ENTERPRISES BE LIABLE
 * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* protocol.c - proxynet protocol verbs
 * vix 05may94 [split out of session.c]
 */

#if !defined(lint) && !defined(codecenter)
static char rcsid[] = "$Id: protocol.c,v 1.2 1996/05/16 04:48:56 vixie Exp $";
#endif

#include "externs.h"
#include "defs.h"

static void	signon(void),
		signoff(void);

static verbFunc	command,
		testVerb,
		connVerb,
		listVerb,
		findVerb,
		lstnVerb,
		helpVerb,
		noopVerb,
		quitVerb;

static const char
		testHelp[] = "a\t\ttest sock addr 'a'",
		connHelp[] = "sra\t\tconnect to 'sra'",
		listHelp[] = "\t\tlist connections",
		findHelp[] = "srch a sel\tfind fld 'srch', report fld 'sel'",
		lstnHelp[] = "cla spa\t\tlisten as 'spa', forward to 'cla'",
		helpHelp[] = "[cmd]\t\tshow command syntax summary",
		noopHelp[] = "noop\t\tdo nothing",
		quitHelp[] = "\t\tclose connection";

static struct verb verbs[] = {
	{ "test", testVerb, 1, 1, testHelp },
	{ "conn", connVerb, 1, 1, connHelp },
	{ "list", listVerb, 0, 0, listHelp },
	{ "find", findVerb, 3, 3, findHelp },
	{ "lstn", lstnVerb, 2, 2, lstnHelp },
	{ "help", helpVerb, 0, 1, helpHelp },
	{ "noop", noopVerb, 0, 0, noopHelp },
	{ "quit", quitVerb, 0, 0, quitHelp },
	{ NULL }
};

static char	msgBuf[ARPA_MAXSTR + 50];
static int	quitFlag;

void
protocol() {
	session(signon, signoff, &quitFlag, verbs);
}

static void
signon() {
	int on = 1;

	setsockopt(STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof on);
	sprintf(msgBuf, "Proxy network service (%s) ready on %s\n",
		Version, HostName);
	(void) emit(0220, msgBuf);
}

static void
signoff() {
	struct lstnDbe *ldbe;

	ListEach(&lstnDbeL, ldbe) {
		dbRemove(ldbe->dbe);
	}
}

static int
testVerb(int argc, const char * const *argv) {
	struct sock sock;

	if (!sock_aton(argv[0], &sock)) {
		sprintf(msgBuf, "<%s> is not valid\n", argv[0]);
		return (emit(0501, msgBuf));
	}
	sprintf(msgBuf, "<%s> is <%s>\n", argv[0], sock_ntoa(&sock));
	return (emit(0250, msgBuf));
}

static int
listEach(const char *key, void *ucp) {
	int code = *(int*)ucp;
	struct prox_tup *ent = dbFetch(key);

	if (!ent) {
		syslog(LOG_DEBUG, "cannot fetch %s: %m", key);
	} else {
		char *bufp = msgBuf;

		bufp += sprintf(bufp, "%03o-<", code);
		bufp += sprintf(bufp, "ctl %s ", sock_ntoa(&ent->ctl));
		bufp += sprintf(bufp, "cla %s ", sock_ntoa(&ent->cla));
		bufp += sprintf(bufp, "cpa %s ", sock_ntoa(&ent->cpa));
		bufp += sprintf(bufp, "spa %s ", sock_ntoa(&ent->spa));
		bufp += sprintf(bufp, "sra %s ", sock_ntoa(&ent->sra));
		bufp += sprintf(bufp, "flg 0x%x>", ent->flags);
		puts(msgBuf);	/* includes automatic newline */
	}
	return (0);	/* "keep going" */
}

static int
listVerb(int argc, const char * const *argv) {
	int code = 0250;

	dbForeach(listEach, &code);
	return (emit(0250, "<>\n"));
}

static int
connVerb(int argc, const char * const *argv) {
	struct sock dest;

	if (!sock_aton(argv[0], &dest)) {
		sprintf(msgBuf, "<%s> is not valid\n", argv[0]);
		return (emit(0501, msgBuf));
	}
	if (!sock_isstream(&dest))
		return (emit(0503, "that socket type isn't implemented\n"));
	return (connStream(&dest));
}

/* XXX - should these... */

enum proxFld {noFld, cla, cpa, spa, sra, ctl};

static enum proxFld
atoFld(const char *src) {
	if (!strcasecmp("cla", src))
		return (cla);
	if (!strcasecmp("cpa", src))
		return (cpa);
	if (!strcasecmp("spa", src))
		return (spa);
	if (!strcasecmp("sra", src))
		return (sra);
	if (!strcasecmp("ctl", src))
		return (ctl);
	return (noFld);
}
		
static const struct sock *
getFld(const struct prox_tup *src, enum proxFld fld) {
	switch (fld) {
	case cla: return (&src->cla);
	case cpa: return (&src->cpa);
	case spa: return (&src->spa);
	case sra: return (&src->sra);
	case ctl: return (&src->ctl);
	default: return (NULL);
	}
	/*NOTREACHED*/
}

/* ...be moved to database.c? */

struct findCtx {
	struct sock srchAddr, selAddr;
	enum proxFld srchFld, selFld;
	int done;
};

static int
findEach(const char *key, void *ucp) {
	struct findCtx *ctx = (struct findCtx *)ucp;
	struct prox_tup *ent = dbFetch(key);
	const struct sock *srch, *sel;

#ifdef DEBUG
	if (ctx->done) {
		syslog(LOG_DEBUG, "findEach: already done");
		return (1);
	}
#endif

	if (!ent) {
		syslog(LOG_DEBUG, "cannot fetch %s: %m", key);
	} else if (!(srch = getFld(ent, ctx->srchFld))) {
		syslog(LOG_WARNING, "bad srchFld (internal error)");
	} else if (!(sel = getFld(ent, ctx->selFld))) {
		syslog(LOG_WARNING, "bad selFld (internal error)");
	} else {
		if (sock_same(srch, &ctx->srchAddr)) {
			ctx->selAddr = *sel;
			ctx->done = TRUE;
			return (1);	/* "stop here" */
		}
	}
	return (0);	/* "keep going" */
}


static int
findVerb(int argc, const char * const *argv) {
	struct findCtx ctx;
	time_t mtime;

	if (dbMtime(&mtime) == 0) {
		time_t now;

		(void) time(&now);
		if ((((u_long)now) - ((u_long)mtime)) < 2)
			sleep(1);
	}

	memset(&ctx, 0, sizeof ctx);
	if ((ctx.srchFld = atoFld(argv[0])) == noFld) {
		sprintf(msgBuf, "cannot search for <%s> - unknown\n", argv[0]);
		return (emit(0503, msgBuf));
	}
	if (!sock_aton(argv[1], &ctx.srchAddr)) {
		sprintf(msgBuf, "<%s> is not valid\n", argv[1]);
		return (emit(0501, msgBuf));
	}
	if ((ctx.selFld = atoFld(argv[2])) == noFld) {
		sprintf(msgBuf, "cannot select <%s> - unknown\n", argv[2]);
		return (emit(0504, msgBuf));
	}
	ctx.done = FALSE;
	dbForeach(findEach, &ctx);
	if (ctx.done) {
		sprintf(msgBuf, "%s <%s>\n", argv[2], sock_ntoa(&ctx.selAddr));
		return (emit(0250, msgBuf));
	}
	sprintf(msgBuf, "%s <%s> not found\n",
		argv[0], sock_ntoa(&ctx.srchAddr));
	return (emit(0540, msgBuf));
}


static int
lstnVerb(int argc, const char * const *argv) {
	struct sock s_cla, s_spa;

	if (!sock_aton(argv[0], &s_cla)) {
		sprintf(msgBuf, "cla <%s> is not valid\n", argv[0]);
		return (emit(0501, msgBuf));
	}
	if (!sock_aton(argv[1], &s_spa)) {
		sprintf(msgBuf, "spa <%s> is not valid\n", argv[0]);
		return (emit(0501, msgBuf));
	}
	if (!sock_isstream(&s_cla) || !sock_isstream(&s_spa))
		return (emit(0503, "only stream sockets can be LSTN'ed\n"));
	return (lstnStream(&s_cla, &s_spa));
}


static int
helpVerb(int argc, const char * const *argv) {
	const struct verb *verb;
	const int code = 0250;

	for (verb = verbs;  verb->verb;  verb++) {
		char sep;

		if (argc && strcasecmp(argv[0], verb->verb))
			continue;
		sep = (!argc && verb[1].verb) ? '-' : ' ';
		printf("%03o%c\t%s\t%s\n", code, sep, verb->verb, verb->help);
	}
	return (code);
}


static int
noopVerb(int argc, const char * const *argv) {
	return (emit(0250, "Ok\n"));
}


static int
quitVerb(int argc, const char * const *argv) {
	quitFlag = TRUE;
	return (emit(0250, "Goodbye\n"));
}
